#include <amsmf/ParseSMF.h>
#include <amsmf/Assert.h>
#include <amt/Endian.h>
#include <algorithm>

namespace AMSMF
{

namespace Parsers
{

// ----------------------------------------------------------------------
// TrackEvent
//

U8 *TrackEvent::get_bytes_(U32 nb)
{
	return nb > MAX_BYTES ? new U8 [nb] : m_regular_bytes;
}

void TrackEvent::dispose_bytes_()
{
	if (m_bytes != m_regular_bytes) delete [] m_bytes;
}

void TrackEvent::resize_bytes_(U32 nb)
{
	if (nb > m_num_bytes_reserved) {
		this->dispose_bytes_();
		m_bytes = this->get_bytes_(nb);
		m_num_bytes_reserved = nb;
	}
	m_num_bytes = nb;
}

TrackEvent::~TrackEvent()
{
	this->dispose_bytes_();
}

TrackEvent::TrackEvent(Type t, U32 d, U32 nb)
	: m_type(t), m_delta(d), m_num_bytes(nb), m_bytes(this->get_bytes_(nb))
{
}

TrackEvent::TrackEvent(const TrackEvent &src)
	: m_type(src.type()), m_delta(src.delta()),
	  m_num_bytes(src.num_bytes()),
	  m_bytes(this->get_bytes_(src.num_bytes()))
{
	memcpy(this->bytes(), src.bytes(), m_num_bytes);
}

TrackEvent &
TrackEvent::operator = (const TrackEvent &src)
{
	m_type = src.type();
	m_delta = src.delta();
	this->resize_bytes_(src.num_bytes());
	memcpy(this->bytes(), src.bytes(), m_num_bytes);
	return *this;
}

void
TrackEvent::reset(Type t, U32 d, U32 nb)
{
	m_type = t;
	m_delta = d;
	m_is_running_status = false;
	this->resize_bytes_(nb);
}

} // Parsers

using namespace Parsers;

namespace
{

typedef FMT::SMF::Endian::Convert<AMT::Endian::ExtToHost> Convert;

class FetchFixedBlock
{
	void *m_buffer;
	U32 m_buffer_length;
	void *m_curptr;
	U32 m_remain;
public:
	~FetchFixedBlock() { }
	FetchFixedBlock() { }
	FetchFixedBlock(void *buf, U32 len)
		: m_buffer(buf), m_buffer_length(len),
		  m_curptr(buf), m_remain(len)
	{ }
	void reset(void *buf, U32 len)
	{
		m_buffer = m_curptr = buf;
		m_buffer_length = m_remain = len;
	}
	bool operator () (InputStream &is)
	{
		U32 len;
		InputStream::Result res;

		while ((len = m_remain) &&
		       (res = is.read(m_curptr,
				      &len)) == InputStream::R_SUCCESS) {
			m_curptr = (U8 *)m_curptr + len;
			m_remain -= len;
		}
		if (res == InputStream::R_EOF)
			throw Ex::UnexpectedFileEnd();
		return res != InputStream::R_SUCCESS;
	}
};

class FetchVarInteger
{
	static const int UNIT_SHIFT = FMT::SMF::VarInteger::UNIT_SHIFT;
	static const U8 CONT_MASK = FMT::SMF::VarInteger::CONT_MASK;
	static const U8 VAL_MASK = FMT::SMF::VarInteger::VAL_MASK;
	U32 m_value;
	U32 m_max_length;
	U32 m_bytes_read;
public:
	FetchVarInteger() : m_value(), m_max_length(), m_bytes_read() { }
	void reset(U32 maxlen)
	{ m_value = 0; m_max_length = maxlen; m_bytes_read = 0; }
	AMT_ACSOR_R(U32, value);
	AMT_ACSOR_R(U32, bytes_read);
	bool operator () (InputStream &is)
	{
		U8 ch;
		U32 len;
		InputStream::Result res;

		for (;;) {
			if (m_bytes_read >= m_max_length)
				throw Ex::ChunkOverrun();
			if (m_bytes_read >= FMT::SMF::VarInteger::MAX_LENGTH)
				throw Ex::TooLongVarInteger();
			len = 1;
			res = is.read(&ch, &len);
			switch (res) {
			case InputStream::R_SUCCESS:
				m_value =
				    (m_value << UNIT_SHIFT) | (ch & VAL_MASK);
				m_bytes_read++;
				if ((ch & CONT_MASK) == 0)
					return false;
				break;
			case InputStream::R_EOF:
				throw Ex::UnexpectedFileEnd();
			default:
				return true;
			}
		}
	}
};

class FetchChunkHeader
{
	static const U32 LEN_TOTAL = FMT::SMF::Chunk::TotalF::LENGTH;
private:
	char m_buffer[LEN_TOTAL];
	FetchFixedBlock m_fetch;
	FMT::SMF::Chunk::MagicF m_magic;
	FMT::SMF::Chunk::LengthF m_length;
public:
	FetchChunkHeader() { }
	void reset() { m_fetch.reset(m_buffer, LEN_TOTAL); }
	bool operator () (InputStream &is)
	{
		if (m_fetch(is))
			return true;
		Convert::convert(&m_magic, m_buffer);
		Convert::convert(&m_length, m_buffer);
		return false;
	}
	AMT_PROXY_ACSOR_RW(U32, magic, magic, value);
	AMT_PROXY_ACSOR_RW(U32, length, length, value);
};

class FetchMetaHeader
{
public:
	static const U32 LEN_TOTAL=(U32)FMT::SMF::Meta::TotalF::LENGTH;
private:
	U8 m_buffer[LEN_TOTAL];
	FetchFixedBlock m_fetch;
public:
	FetchMetaHeader() : m_fetch(m_buffer, LEN_TOTAL) { }
	void reset() { m_fetch.reset(m_buffer, LEN_TOTAL); }
	bool operator () (InputStream &is)
	{
		return m_fetch(is);
	}
	U8 kind() const
	{ return Convert::fetch<FMT::SMF::Meta::KindF>(m_buffer); }
	U8 length() const
	{ return Convert::fetch<FMT::SMF::Meta::LengthF>(m_buffer); }
};

class FetchSMFHeader
{
public:
	static const U32 LEN_TOTAL = (U32)FMT::SMF::Header::TotalF::LENGTH;
private:
	U8 m_buffer[LEN_TOTAL];
	FetchFixedBlock m_fetch;
	SMFHeader &m_header;
public:
	FetchSMFHeader(SMFHeader &hdr)
		: m_fetch(m_buffer, LEN_TOTAL), m_header(hdr) { }
	void reset() { m_fetch.reset(m_buffer, LEN_TOTAL); }
	bool operator () (InputStream &is)
	{
		if (m_fetch(is))
			return true;
		m_header.format() =
		    Convert::fetch<FMT::SMF::Header::FormatF>(m_buffer);
		m_header.num_tracks() =
		    Convert::fetch<FMT::SMF::Header::NumTracksF>(m_buffer);
		m_header.division() =
		    Convert::fetch<FMT::SMF::Header::DivisionF>(m_buffer);
		return false;
	}
	AMT_ACSOR_R(SMFHeader, header);
};

class FetchExtra
{
	static const U32 LEN_BUF = 256;
	Hooks *m_hooks;
	U32 m_remain;
public:
	FetchExtra() { }
	void reset(Hooks &h, U32 r) { m_hooks = &h, m_remain = r; }
	bool operator () (InputStream &is)
	{
		InputStream::Result res;
		U32 len;
		U8 buf[LEN_BUF];

		while ((len = std::min(m_remain, LEN_BUF)) &&
		       (res = is.read(buf, &len)) == InputStream::R_SUCCESS) {
			m_hooks->header_extra_bytes(buf, len);
			m_remain -= len;
		}
		if (res == InputStream::R_EOF)
			throw Ex::UnexpectedFileEnd();
		return res != InputStream::R_SUCCESS;
	}
};

const U32 FetchExtra::LEN_BUF;

inline bool
peek_byte(InputStream &is, U8 *byte)
{
	switch (is.peek(byte)) {
	case InputStream::R_EOF:
		throw Ex::UnexpectedFileEnd();
	case InputStream::R_AGAIN:
		return true;
	default:
		;
	}
	return false;
}

inline bool
read_byte(InputStream &is, U8 *byte)
{
	U32 len = 1;
	switch (is.read(byte, &len)) {
	case InputStream::R_EOF:
		throw Ex::UnexpectedFileEnd();
	case InputStream::R_AGAIN:
		return true;
	default:
		;
	}
	return false;
}


class StateMachine;
class MachineStack
{
	InputStream &m_stream;
	Hooks &m_hooks;
	StateMachine *m_top;
public:
	void push(StateMachine *);
	void pop();
	MachineStack(InputStream &, Hooks &, StateMachine *);
	~MachineStack();
	bool parse();
	AMT_REF_ACSOR_CONST(InputStream, stream);
	AMT_REF_ACSOR_CONST(Hooks, hooks);
};

class StateMachine
{
	friend class MachineStack;
	AMT_NOCOPY(StateMachine);
private:
	StateMachine *m_upper;
	MachineStack *m_stack;
	//
	void push_(StateMachine *up)
	{
		m_upper = up;
		m_stack = &up->stack();
	}
	StateMachine *pop_() { return AMT::renounce(&m_upper); }
protected:
	AMT_PTR_TO_REF_ACSOR_CONST(MachineStack, stack);
	InputStream &stream() const { return m_stack->stream(); }
	Hooks &hooks() const { return m_stack->hooks(); }
public:
	enum Result
	{
		R_DONE,
		R_SUSPEND,
		R_STACK,
	};
	virtual ~StateMachine() = 0;
	virtual Result parse() = 0;
	StateMachine() : m_upper(), m_stack() { }
};
AMT_DEF_INLINE_DTOR(StateMachine);


inline void
MachineStack::push(StateMachine *newmac)
{
	AMSMF_ASSERT_I_("push", m_top != NULL);
	newmac->push_(m_top);
	m_top = newmac;
}

inline void
MachineStack::pop()
{
	if (m_top)
		delete AMT::renounce(&m_top, m_top->pop_());
}

MachineStack::MachineStack(InputStream &is, Hooks &h, StateMachine *top)
	: m_stream(is), m_hooks(h), m_top(top)
{
	top->m_upper = NULL;
	top->m_stack = this;
}

MachineStack::~MachineStack()
{
	while (m_top) this->pop();
}

bool
MachineStack::parse()
{
	StateMachine::Result rv;

	while (m_top) {
		rv = m_top->parse();
		switch (rv) {
		case StateMachine::R_DONE:
			this->pop();
			break;
		case StateMachine::R_STACK:
			break;
		case StateMachine::R_SUSPEND:
			return true;
		}
	}
	return false;
}



class EventContext
{
	TrackEvent m_event;
	U32 m_curpos;
	U32 m_remain;
public:
	typedef TrackEvent::Type Type;
	EventContext() : m_remain(0) { }
	void reset(Type t, U32 d, U32 nb)
	{
		m_curpos = 0;
		m_remain = nb;
		if (t == TrackEvent::T_META) {
			m_event.reset(t, d, nb+1);
			m_event.bytes()[nb] = '\0';
		} else {
			m_event.reset(t, d, nb);
		}
	}
	AMT_ACSOR_RW(TrackEvent, event);
	AMT_ACSOR_RW(U32, remain);
	AMT_ACSOR_RW(U32, curpos);
	void ensure_bytes(U32 len)
	{
		AMSMF_ASSERT_I_("ensure_bytes", m_remain >= len);
	}
	void push_byte(U8 byte)
	{
		this->ensure_bytes(1);
		m_event.bytes()[m_curpos] = byte;
		m_curpos++;
		m_remain--;
	}
	void push_bytes(U8 *bytes, U32 len)
	{
		this->ensure_bytes(len);
		memcpy(&m_event.bytes()[m_curpos], bytes, len);
		m_curpos += len;
		m_remain -= len;
	}
	U8 *curptr() { return &m_event.bytes()[m_curpos]; }
};

class TrackContext
{
	int m_track_no;
	U32 m_remain_bytes;
	U8 m_last_midi_status;
public:
	TrackContext()
		: m_track_no(0), m_remain_bytes(0), m_last_midi_status(0) { }
	AMT_ACSOR_RW(int, track_no);
	AMT_ACSOR_RW(U32, remain_bytes);
	AMT_ACSOR_RW(U8, last_midi_status);
	bool is_last_midi_status_valid() const
	{ return m_last_midi_status != 0; }
	void invalidate_last_midi_status() { m_last_midi_status = 0; }
	void ensure_remain_bytes(U32 size)
	{
		if (m_remain_bytes < size)
			throw Ex::ChunkOverrun();
	}
	void decr_remain_bytes(U32 decr)
	{
		this->ensure_remain_bytes(decr);
		m_remain_bytes -= decr;
	}
};

class SMFContext
{
	SMFHeader m_header;
	int m_current_track;
public:
	SMFContext() : m_current_track(0) { }
	AMT_ACSOR_RW(SMFHeader, header);
	AMT_ACSOR_RW(int, current_track);
};


#define BEGIN_STATES(n)							      \
switch ((m_state)) {							      \
case n:
#define SHIFT_STATE(n)							      \
	(m_state) = (n);						      \
	/*FALLTHROUGH*/							      \
case n:
#define END_STATES							      \
}									      \
return R_DONE
#define SUSPEND_STATE							      \
return R_SUSPEND
#define PUSH_MACHINE(newm, nextp)					      \
this->stack().push((newm)); (m_state) = (nextp); return R_STACK


class SMFTrackDataMachine : public StateMachine
{
public:
	enum State
	{
		S_INIT,
		S_BEGIN_DELTA,
		S_DELTA,
		S_EVENT_STATUS,
		S_EVENT_META,
		S_EVENT_EXCL,
		S_EVENT_ARBIT,
		S_EVENT_BODY_INIT,
		S_EVENT_BODY,
		S_DONE
	};
private:
	State m_state;
	TrackContext &m_track;
	EventContext m_event;
	FetchFixedBlock m_fetch_block;
	FetchVarInteger m_fetch_varint;
	FetchMetaHeader m_fetch_meta;
	U32 m_delta;
public:
	~SMFTrackDataMachine() { }
	SMFTrackDataMachine(TrackContext &tr)
		: m_state(S_INIT), m_track(tr) { }
	Result parse();
};

SMFTrackDataMachine::Result
SMFTrackDataMachine::parse()
{
	BEGIN_STATES(S_INIT) {
		this->hooks().begin_track(m_track.track_no(),
					  m_track.remain_bytes());
	}
begin:
	SHIFT_STATE(S_BEGIN_DELTA) {
		if (m_track.remain_bytes() == 0)
			goto done;
		m_fetch_varint.reset(m_track.remain_bytes());
	}
	SHIFT_STATE(S_DELTA) {
		if (m_fetch_varint(this->stream()))
			return R_SUSPEND;
		m_track.decr_remain_bytes(m_fetch_varint.bytes_read());
		m_delta = m_fetch_varint.value();
		m_track.ensure_remain_bytes(1);
	}
	SHIFT_STATE(S_EVENT_STATUS) {
		U8 byte;
		if (peek_byte(this->stream(), &byte))
			return R_SUSPEND;
		bool iss = FMT::SMF::MSG::is_status_byte(byte);
		if (iss) {
			m_track.decr_remain_bytes(1);
			if (read_byte(this->stream(), &byte))
				return R_SUSPEND;
		} else {
			if (!m_track.is_last_midi_status_valid())
				throw Ex::IllegalRunningStatus();
			byte = m_track.last_midi_status();
		}

		switch (byte) {
		case FMT::SMF::MSG::SMFMeta::STATUS:
			m_track.decr_remain_bytes(FetchMetaHeader::LEN_TOTAL);
			m_fetch_meta.reset();
			goto event_meta;
		case FMT::SMF::MSG::SMFSysEx::STATUS:
			m_fetch_varint.reset(m_track.remain_bytes());
			goto event_excl;
		case FMT::SMF::MSG::SMFArbitBytes::STATUS:
			m_fetch_varint.reset(m_track.remain_bytes());
			goto event_arbit;
		}
		if (!FMT::SMF::MSG::is_ch_msg(byte))
			throw Ex::IllegalSystemMessage();
		int len = FMT::SMF::MSG::lookup_msg_length(byte);
		if (len == 0)
			throw Ex::IllegalStatusByte();
		m_event.reset(TrackEvent::T_MIDI, m_delta, len);
		m_event.event().is_running_status() = !iss;
		m_event.push_byte(byte);
		m_track.last_midi_status() = byte;
		goto body;
	}
event_meta:
	SHIFT_STATE(S_EVENT_META) {
		if (m_fetch_meta(this->stream()))
			return R_SUSPEND;
		m_event.reset(TrackEvent::T_META,
			      m_delta, m_fetch_meta.length()+3);
		m_event.push_byte(FMT::SMF::MSG::SMFMeta::STATUS);
		m_event.push_byte(m_fetch_meta.kind());
		m_event.push_byte(m_fetch_meta.length());
		goto body;
	}
event_excl:
	SHIFT_STATE(S_EVENT_EXCL) {
		if (m_fetch_varint(this->stream()))
			return R_SUSPEND;
		m_track.decr_remain_bytes(m_fetch_varint.bytes_read());
		m_event.reset(TrackEvent::T_SYSEX,
			      m_delta, m_fetch_varint.value()+1);
		m_event.push_byte(FMT::SMF::MSG::SysEx::STATUS);
		goto body;
	}
event_arbit:
	SHIFT_STATE(S_EVENT_ARBIT) {
		if (m_fetch_varint(this->stream()))
			return R_SUSPEND;
		m_track.decr_remain_bytes(m_fetch_varint.bytes_read());
		m_event.reset(TrackEvent::T_SYSEX,
			      m_delta, m_fetch_varint.value());
		goto body;
	}
body:
	SHIFT_STATE(S_EVENT_BODY_INIT) {
		m_track.decr_remain_bytes(m_event.remain());
		m_fetch_block.reset(m_event.curptr(), m_event.remain());
	}
	SHIFT_STATE(S_EVENT_BODY) {
		if (m_fetch_block(this->stream()))
			return R_SUSPEND;
		this->hooks().track_event(m_track.track_no(), m_event.event());
	}
done:
	SHIFT_STATE(S_DONE) {
		if (m_track.remain_bytes() > 0)
			goto begin;
		this->hooks().end_track(m_track.track_no());
	}
	END_STATES;
}

class SMFTrackMachine : public StateMachine
{
public:
	enum State
	{
		S_INIT,
		S_HEADER,
		S_DONE
	};
private:
	State m_state;
	SMFContext &m_smf;
	FetchChunkHeader m_fetch;
	TrackContext m_track;
public:
	~SMFTrackMachine() { }
	SMFTrackMachine(SMFContext &ctx) : m_state(S_INIT), m_smf(ctx) { }
	Result parse();
};

SMFTrackMachine::Result
SMFTrackMachine::parse()
{
	BEGIN_STATES(S_INIT) {
		m_fetch.reset();
	}
	SHIFT_STATE(S_HEADER) {
		if (m_fetch(this->stream()))
			return R_SUSPEND;
		if (m_fetch.magic() != FMT::SMF::Track::MAGIC)
			throw Ex::IllegalMagic();
		m_track.track_no() = m_smf.current_track()++;
		m_track.remain_bytes() = m_fetch.length();
		m_track.invalidate_last_midi_status();
		PUSH_MACHINE(new SMFTrackDataMachine(m_track), S_DONE);
	}
	SHIFT_STATE(S_DONE) {
	}
	END_STATES;
}


class SMFHeaderMachine : public StateMachine
{
	static const U32 MIN_SIZE = FMT::SMF::Header::TotalF::LENGTH;
public:
	enum State
	{
		S_INIT,
		S_CHUNK_HEADER,
		S_SMF_INFO,
		S_EXTRA_BYTES,
		S_DONE,
	};
private:
	State m_state;
	SMFContext &m_smf;
	FetchChunkHeader m_fetch_chunk_header;
	FetchSMFHeader m_fetch_smf_info;
	FetchExtra m_fetch_extra;
public:
	~SMFHeaderMachine() { }
	SMFHeaderMachine(SMFContext &ctx)
		: m_state(S_INIT), m_smf(ctx), m_fetch_smf_info(ctx.header())
	{ }
	Result parse();
};

SMFHeaderMachine::Result
SMFHeaderMachine::parse()
{
	BEGIN_STATES(S_INIT) {
		m_fetch_chunk_header.reset();
	}
	SHIFT_STATE(S_CHUNK_HEADER) {
		if (m_fetch_chunk_header(this->stream()))
			return R_SUSPEND;
		if (m_fetch_chunk_header.magic() != FMT::SMF::Header::MAGIC)
			throw Ex::IllegalMagic();
		if (m_fetch_chunk_header.length() < FetchSMFHeader::LEN_TOTAL)
			throw Ex::TooShortHeader();
		m_fetch_smf_info.reset();
		m_smf.header().total_header_size() =
		    m_fetch_chunk_header.length();
		m_smf.header().num_extra_bytes() =
		    m_fetch_chunk_header.length() - FetchSMFHeader::LEN_TOTAL;
	}
	SHIFT_STATE(S_SMF_INFO) {
		if (m_fetch_smf_info(this->stream()))
			return R_SUSPEND;
		this->hooks().begin_smf(m_smf.header());
		if (m_smf.header().num_extra_bytes() == 0)
			goto done;
		m_fetch_extra.reset(
			this->hooks(), m_smf.header().num_extra_bytes());
		this->hooks().begin_header_extra_bytes(
			m_smf.header().num_extra_bytes());
	}
	SHIFT_STATE(S_EXTRA_BYTES) {
		if (m_fetch_extra(this->stream()))
			return R_SUSPEND;
		this->hooks().end_header_extra_bytes();
	}
done:
	SHIFT_STATE(S_DONE) {
	}
	END_STATES;
}

class SMFMachine : public StateMachine
{
public:
	enum State
	{
		S_INIT,
		S_SMF_HEADER,
		S_SMF_TRACKS,
		S_DONE
	};
private:
	State m_state;
	SMFContext m_smf;
public:
	~SMFMachine() { }
	SMFMachine() : m_state(S_INIT) { }
	Result parse();
};

SMFMachine::Result
SMFMachine::parse()
{
	BEGIN_STATES(S_INIT) {
	}
	SHIFT_STATE(S_SMF_HEADER) {
		PUSH_MACHINE(new SMFHeaderMachine(m_smf), S_SMF_TRACKS);
	}
	SHIFT_STATE(S_SMF_TRACKS) {
		if (m_smf.current_track() < m_smf.header().num_tracks()) {
			PUSH_MACHINE(new SMFTrackMachine(m_smf),
				     S_SMF_TRACKS);
		}
	}
	SHIFT_STATE(S_DONE) {
		this->hooks().end_smf();
	}
	END_STATES;
}

}

// ----------------------------------------------------------------------
//

class Parser::Impl_
{
	MachineStack m_stack;
public:
	Impl_(InputStream &is, Hooks &h) : m_stack(is, h, new SMFMachine) { }
	bool parse() { return m_stack.parse(); }
};

Parser::~Parser() { }

Parser::Parser(InputStream &is, Hooks &h) : m_impl(new Impl_(is, h)) { }

bool Parser::parse() { return m_impl->parse(); }


// ----------------------------------------------------------------------
//

const char *
TrackEvent::type_name() const
{
	switch (m_type) {
	case T_MIDI:
		return "MIDI";
	case T_SYSEX:
		return "SysEx";
	case T_META:
		return "META";
	}
	return "Unknown";
}


}
