#ifndef AMSMF_PARSE_SMF_H_
#define AMSMF_PARSE_SMF_H_

#include <amsmf/Elements.h>
#include <amsmf/Exception.h>
#include <amt/ScopedPtr.h>

namespace AMSMF
{

class Parser;

namespace Parsers
{

class TrackEvent;
class InputStream;
class Hooks;

class TrackEvent
{
public:
	static const int MAX_BYTES = 4;
	enum Type
	{
		T_MIDI, T_SYSEX, T_META
	};
private:
	Type m_type;
	U32 m_delta;
	U32 m_num_bytes;
	U32 m_num_bytes_reserved;
	U8 *m_bytes;
	U8 m_regular_bytes[MAX_BYTES];
	bool m_is_running_status;
	//
	U8 *get_bytes_(U32);
	void dispose_bytes_();
	void resize_bytes_(U32);
public:
	~TrackEvent();
	TrackEvent()
		: m_type(T_MIDI), m_delta(0), m_num_bytes(0),
		  m_num_bytes_reserved(0), m_bytes(NULL),
		  m_is_running_status(false)
	{ }
	TrackEvent(Type, U32, U32);
	TrackEvent(const TrackEvent &);
	TrackEvent &operator = (const TrackEvent &);
	void reset(Type, U32, U32);
	AMT_ACSOR_RW(Type, type);
	AMT_ACSOR_RW(U32, delta);
	AMT_ACSOR_R(U32, num_bytes);
	AMT_ACSOR_RW(bool, is_running_status);
	U8 *bytes() { return m_bytes; }
	const U8 *bytes() const { return m_bytes; }
	template <const Type TYPE_>
	bool is_type() const { return m_type == TYPE_; }
	bool is_midi() const { return this->is_type<T_MIDI>(); }
	bool is_sysex() const { return this->is_type<T_SYSEX>(); }
	bool is_meta() const { return this->is_type<T_META>(); }
	const char *type_name() const;
};

class InputStream
{
public:
	enum Result
	{
		R_SUCCESS=0,
		R_AGAIN,
		R_EOF,
	};
	virtual ~InputStream() = 0;
	virtual Result read(void *, U32 *) = 0;
	virtual Result peek(U8 *) = 0;
};
AMT_DEF_INLINE_DTOR(InputStream);

class Hooks
{
public:
	typedef class SMFHeader SMFHeader;
	typedef class TrackEvent TrackEvent;
	virtual ~Hooks() = 0;
	virtual void begin_smf(const SMFHeader &) = 0;
	virtual void end_smf() = 0;
	virtual void begin_header_extra_bytes(U32) = 0;
	virtual void end_header_extra_bytes() = 0;
	virtual void header_extra_bytes(const U8 *, U32) = 0;
	virtual void begin_track(int, U32) = 0;
	virtual void end_track(int) = 0;
	virtual void track_event(int, const TrackEvent &) = 0;
};
AMT_DEF_INLINE_DTOR(Hooks);

namespace MSG_
{
using namespace FMT::SMF::MSG;

template <const TrackEvent::Type TYPE_, class Msg_>
class ReaderDef_ : public Reader<Msg_, ReaderDef_<TYPE_, Msg_> >::Def
{
	friend class ReaderBase<Msg_, ReaderDef_<TYPE_, Msg_> >;
	const TrackEvent &m_event;
public:
	static bool is_match(const TrackEvent &ev)
	{
		AMSMF_ASSERT_U_("is_match", ev.num_bytes() > 0);
		return ev.is_type<TYPE_>() && Msg_::is_match(ev.bytes());
	}
protected:
	const TrackEvent &event_() const { return m_event; }
	const U8 *bytes_() const { return m_event.bytes(); }
public:
	ReaderDef_(const TrackEvent &ev) : m_event(ev) { }
	U32 delta() const { return m_event.delta(); }
};

template <class Msg_>
struct MIDIReader
{
	typedef ReaderDef_<TrackEvent::T_MIDI, Msg_> Def;
};

template <class Msg_>
struct SysExReader
{
	struct Def : public ReaderDef_<TrackEvent::T_SYSEX, Msg_>
	{
		Def(const TrackEvent &ev)
			: ReaderDef_<TrackEvent::T_SYSEX, Msg_>(ev)
		{ }
		U32 length() const { return this->event_().num_bytes(); }
		const U8 *bytes() const { return this->bytes_(); }
	};
};

template <class Msg_>
struct MetaReader
{
	typedef ReaderDef_<TrackEvent::T_META, Msg_> Def;
};

}

typedef MSG_::MIDIReader<MSG_::ChNoteOff>::Def ChNoteOff;
typedef MSG_::MIDIReader<MSG_::ChNoteOn>::Def ChNoteOn;
typedef MSG_::MIDIReader<MSG_::ChPolyKeyPressure>::Def ChPolyKeyPressure;
typedef MSG_::MIDIReader<MSG_::ChControlChange>::Def ChControlChange;
typedef MSG_::MIDIReader<MSG_::ChProgramChange>::Def ChProgramChange;
typedef MSG_::MIDIReader<MSG_::ChChannelPressure>::Def ChChannelPressure;
typedef MSG_::MIDIReader<MSG_::ChPitchBend>::Def ChPitchBend;

// note that channel messages are only allowed by SMF spec,
// except for SysEx events.

typedef MSG_::SysExReader<MSG_::SMFSysEx>::Def SMFSysEx;
typedef MSG_::SysExReader<MSG_::SMFArbitBytes>::Def SMFArbitBytes;

typedef MSG_::MetaReader<MSG_::SMFMetaSeqNo>::Def SMFMetaSeqNo;
typedef MSG_::MetaReader<MSG_::SMFMetaText>::Def SMFMetaText;
typedef MSG_::MetaReader<MSG_::SMFMetaCopyright>::Def SMFMetaCopyright;
typedef MSG_::MetaReader<MSG_::SMFMetaSeqTrackName>::Def SMFMetaSeqTrackName;
typedef MSG_::MetaReader<MSG_::SMFMetaInstName>::Def SMFMetaInstName;
typedef MSG_::MetaReader<MSG_::SMFMetaLyric>::Def SMFMetaLyric;
typedef MSG_::MetaReader<MSG_::SMFMetaMarker>::Def SMFMetaMarker;
typedef MSG_::MetaReader<MSG_::SMFMetaQuePoint>::Def SMFMetaQuePoint;
typedef MSG_::MetaReader<MSG_::SMFMetaChPrefix>::Def SMFMetaChPrefix;
typedef MSG_::MetaReader<MSG_::SMFMetaEndOfTrack>::Def SMFMetaEndOfTrack;
typedef MSG_::MetaReader<MSG_::SMFMetaSetTempo>::Def SMFMetaSetTempo;
typedef MSG_::MetaReader<MSG_::SMFMetaSMPTEOffset>::Def SMFMetaSMPTEOffset;
typedef MSG_::MetaReader<MSG_::SMFMetaTimeSign>::Def SMFMetaTimeSign;
typedef MSG_::MetaReader<MSG_::SMFMetaKeySign>::Def SMFMetaKeySign;

}

class Parser
{
private:
	class Impl_;
	AMT::Ptr::Scoped<Impl_> m_impl;
public:
	~Parser();
	Parser(Parsers::InputStream &, Parsers::Hooks &);
	bool parse();
};

}

#endif
