#ifndef AMT_ENDIAN_H_
#define AMT_ENDIAN_H_

#include <amt/Types.h>
#ifdef __NetBSD__
#include <sys/endian.h>
#endif
#include <cstring>

namespace AMT
{


namespace Endian
{

struct HostToExt { };
struct ExtToHost { };

template <const Size ofs>
struct Offset
{
	static const Size OFFSET = ofs;
	static const Size LENGTH = 0;
};

template <typename T, class Prev = Offset<0> >
class Field
{
public:
	typedef T ValueType;
	static const Size OFFSET = Prev::OFFSET + Prev::LENGTH;
	static const Size LENGTH = sizeof (T);
private:
	ValueType m_value;
public:
	Field() { }
	Field(ValueType v) : m_value(v) { }
	AMT_ACSOR_RW(ValueType, value);
	static Size offset() { return OFFSET; }
	static Size length() { return LENGTH; }
};

template <class Prev>
class Total
{
public:
	static const Size OFFSET = 0;
	static const Size LENGTH = Prev::OFFSET + Prev::LENGTH;
};

template <class Inh_>
struct Base_
{
	struct To
	{
		template <typename Int_>
		static void convert(const Int_ *v, void *ptr, Size ofs = 0)
		{
			Int_ cv=Inh_::template to<Int_>(*v);
			std::memcpy((char *)ptr + ofs, &cv, sizeof (cv));
		}
		static void convert(const U8 *v, void *ptr, Size ofs = 0)
		{
			*((char *)ptr + ofs) = *v;
		}
		template <typename Int_>
		static Int_ convert(Int_ v)
		{
			return Inh_::template to<Int_>(v);
		}
		template <typename Int_, class Prev_>
		static void convert(const Field<Int_, Prev_> *v, void *ptr)
		{
			To::template convert<Int_>(
				&v->value(), ptr, v->offset());
		}
		template <class Field_>
		static void store(void *ptr, typename Field_::ValueType v)
		{
			typedef typename Field_::ValueType Int_;
			To::template convert<Int_>(&v, ptr, Field_::OFFSET);
		}
	};
	struct From
	{
		template <typename Int_>
		static void convert(Int_ *v, const void *ptr, Size ofs=0)
		{
			std::memcpy(v, (const char *)ptr + ofs, sizeof (*v));
			*v = Inh_::template from<Int_>(*v);
		}
		static void convert(U8 *v, const void *ptr, Size ofs = 0)
		{
			*v = *((const char *)ptr + ofs);
		}
		template <typename Int_>
		static Int_ convert(Int_ v)
		{
			return Inh_::template from<Int_>(*v);
		}
		template <typename Int_, class Prev_>
		static void convert(Field<Int_, Prev_> *v, const void *ptr)
		{
			From::template convert<Int_>(
				&v->value(), ptr, v->offset());
		}
		template <class Field_>
		static typename Field_::ValueType fetch(const void *ptr)
		{
			typedef typename Field_::ValueType Int_;
			Int_ v;
			From::template convert<Int_>(&v, ptr, Field_::OFFSET);
			return v;
		}
	};
};

struct Big : public Base_<Big>
{
	template <typename Int_> static Int_ to(Int_);
	template <typename Int_> static Int_ from(Int_);
	template <typename> struct Convert;
};

#ifdef __NetBSD__
template <> inline U8 Big::to(U8 v) { return v; }
template <> inline U16 Big::to(U16 v) { return htobe16(v); }
template <> inline U32 Big::to(U32 v) { return htobe32(v); }
template <> inline U64 Big::to(U64 v) { return htobe64(v); }
template <> inline U8 Big::from(U8 v) { return v; }
template <> inline U16 Big::from(U16 v) { return be16toh(v); }
template <> inline U32 Big::from(U32 v) { return be32toh(v); }
template <> inline U64 Big::from(U64 v) { return be64toh(v); }
#endif
#ifdef _WIN32
template <> inline U8 Big::to(U8 v) { return v; }
template <> inline U16 Big::to(U16 v)
{ return (v>>8) | (v<<8); }
template <> inline U32 Big::to(U32 v)
{ return Big::to<U16>(v>>16) | ((U32)Big::to<U16>(v)<<16); }
template <> inline U8 Big::from(U8 v) { return v; }
template <> inline U16 Big::from(U16 v) { return Big::to(v); }
template <> inline U32 Big::from(U32 v) { return Big::to(v); }
#endif

template <> struct Big::Convert<HostToExt> : public Big::To { };
template <> struct Big::Convert<ExtToHost> : public Big::From { };

struct Little : public Base_<Little>
{
	template <typename Int_> static Int_ to(Int_);
	template <typename Int_> static Int_ from(Int_);
	template <typename> struct Convert;
};

#ifdef __NetBSD__
template <> inline U8 Little::to(U8 v) { return v; }
template <> inline U16 Little::to(U16 v) { return htole16(v); }
template <> inline U32 Little::to(U32 v) { return htole32(v); }
template <> inline U64 Little::to(U64 v) { return htole64(v); }
template <> inline U8 Little::from(U8 v) { return v; }
template <> inline U16 Little::from(U16 v) { return le16toh(v); }
template <> inline U32 Little::from(U32 v) { return le32toh(v); }
template <> inline U64 Little::from(U64 v) { return le64toh(v); }
#endif
#ifdef _WIN32
template <> inline U8 Little::to(U8 v) { return v; }
template <> inline U16 Little::to(U16 v) { return v; }
template <> inline U32 Little::to(U32 v) { return v; }
template <> inline U64 Little::to(U64 v) { return v; }
template <> inline U8 Little::from(U8 v) { return v; }
template <> inline U16 Little::from(U16 v) { return v; }
template <> inline U32 Little::from(U32 v) { return v; }
template <> inline U64 Little::from(U64 v) { return v; }
#endif

template <> struct Little::Convert<HostToExt> : public Little::To { };
template <> struct Little::Convert<ExtToHost> : public Little::From { };


}

}

#endif
