//
//
//

#ifndef AMT_INTRUSIVE_DOUBLE_LINK_H_
#define AMT_INTRUSIVE_DOUBLE_LINK_H_

#include <amt/Defines.h>
#include <amt/Assert.h>

namespace AMT
{

namespace Predicates
{

struct AlwaysTrue
{
	template <class T_>
	bool operator () (const T_ &) const { return true; }
};

}

namespace IntrusiveDoubleLink
{

template <class, class> class Node;

template <class T_>
struct SimpleAccessor
{
	typedef SimpleAccessor<T_> A_;
	typedef Node<T_, A_> N_;
	typedef typename N_::Container C_;
	static N_ *get_node(T_ *p) { return p; }
	static const N_ *get_node(const T_ *p) { return p; }
	static void dispose(C_ *) { }
};

template <class T_, class A_ = SimpleAccessor<T_> >
class Container
{
	friend class Node<T_, A_>;
	AMT_NOCOPY(Container);
protected:
	AMT_DECL_INNER_FRIEND_CLASS(Access_);
	AMT_DECL_INNER_FRIEND_CLASS(Forward_);
	AMT_DECL_INNER_FRIEND_CLASS(Backward_);
public:
	typedef Node<T_, A_> Node;
private:
	T_ *m_head, *m_tail;
protected:
	class Access_ : public A_
	{
		AMT_BEGIN_CONTRACT(main)
		{
			Node *n AMT_UNUSED = A_::get_node((T_ *)NULL);
			const Node *m AMT_UNUSED =
			    A_::get_node((const T_ *)NULL);
			A_::dispose((Container *)NULL);
		}
	protected:
		static Node *get_node_(T_ *t)
		{
			Node *n = A_::get_node(t);
			AMT_ASSERT_LIBAMT_U_("get_node",
					       !n || !n->is_deadbeef());
			return n;
		}
		static const Node *get_node_(const T_ *t)
		{
			const Node *n = A_::get_node(t);
			AMT_ASSERT_LIBAMT_U_("get_node",
					       !n || !n->is_deadbeef());
			return n;
		}
		static void dispose(Container *c) { A_::dispose(c); }
		static T_ *get_head_(const Container *c) { return c->m_head; }
		static T_ *get_tail_(const Container *c) { return c->m_tail; }
		static T_ **get_head_ref_(Container *c) { return &c->m_head; }
		static T_ **get_tail_ref_(Container *c) { return &c->m_tail; }
		static T_ *get_next_(const T_ *n)
		{ return get_node_(n)->m_next; }
		static T_ *get_prev_(const T_ *n)
		{ return get_node_(n)->m_prev; }
		static T_ **get_next_ref_(Container *c, T_ *n)
		{ return n ? &get_node_(n)->m_next : get_head_ref_(c); }
		static T_ **get_prev_ref_(Container *c, T_ *n)
		{ return n ? &get_node_(n)->m_prev : get_tail_ref_(c); }
		static void set_head_(Container *c, T_ *n) { c->m_head = n; }
		static void set_tail_(Container *c, T_ *n) { c->m_tail = n; }
		static void set_next_(T_ *n, T_ *p)
		{ get_node_(n)->m_next = p; }
		static void set_prev_(T_ *n, T_ *p)
		{ get_node_(n)->m_prev = p; }
		static void clear_node_(T_ *n)
		{
			get_node_(n)->m_next = get_node_(n)->m_prev = NULL;
		}
	public:
		static void confirm_() { AMT_CONFIRM_CONTRACT(main); }
	};
	class Reverse_;
	class Forward_ : public Access_
	{
		friend class Container;
	protected:
		using Access_::get_node_;
		using Access_::get_head_;
		using Access_::get_tail_;
		using Access_::get_head_ref_;
		using Access_::get_tail_ref_;
		using Access_::get_next_;
		using Access_::get_prev_;
		using Access_::get_next_ref_;
		using Access_::get_prev_ref_;
		using Access_::set_head_;
		using Access_::set_tail_;
		using Access_::set_next_;
		using Access_::set_prev_;
		typedef Reverse_ Inverse_;
	};
	class Reverse_ : public Access_
	{
		friend class Container;
	protected:
		static T_ *get_head_(const Container *c)
		{ return Access_::get_tail_(c); }
		static T_ *get_tail_(const Container *c)
		{ return Access_::get_head_(c); }
		static T_ **get_head_ref_(Container *c)
		{ return Access_::get_tail_ref_(c); }
		static T_ **get_tail_ref_(Container *c)
		{ return Access_::get_head_ref_(c); }
		static T_ *get_next_(const T_ *n)
		{ return Access_::get_prev_(n); }
		static T_ *get_prev_(const T_ *n)
		{ return Access_::get_next_(n); }
		static T_ **get_next_ref_(Container *c, T_ *n)
		{ return Access_::get_prev_ref_(c, n); }
		static T_ **get_prev_ref_(Container *c, T_ *n)
		{ return Access_::get_next_ref_(c, n); }
		static void set_head_(Container *c, T_ *n)
		{ Access_::set_tail_(c, n); }
		static void set_tail_(Container *c, T_ *n)
		{ Access_::set_head_(c, n); }
		static void set_next_(T_ *n, T_ *m)
		{ Access_::set_prev_(n, m); }
		static void set_prev_(T_ *n, T_ *m)
		{ Access_::set_next_(n, m); }
		typedef Forward_ Inverse_;
	};
private:
	static void confirm_newnode_(const char *caller, T_ *n)
	{
		AMT_ASSERT_LIBAMT_U_(caller, Forward_::get_next_(n) == NULL);
		AMT_ASSERT_LIBAMT_U_(caller, Forward_::get_prev_(n) == NULL);
	}
protected:
	template <class Ac_, typename Pred_>
	T_ *assumed_next_pred_(T_ *ptr, const Pred_ &p) const
	{
		if (ptr) {
			do {
				ptr = Ac_::get_next_(ptr);
			} while (ptr && !p(*ptr));
		}
		return ptr;
	}
	template <class Ac_, typename Pred_>
	T_ *assumed_prev_pred_(T_ *ptr, const Pred_ &p) const
	{ return assumed_next_pred_<typename Ac_::Inverse_, Pred_>(ptr, p); }
	template <class Ac_, typename Pred_>
	T_ *assumed_head_pred_(const Pred_ &p) const
	{
		T_ *h = Ac_::get_head_(this);
		return (!h || p(*h)) ? h :
		    this->assumed_next_pred_<Ac_, Pred_>(h, p);
	}
	template <class Ac_, typename Pred_>
	T_ *assumed_tail_pred_(const Pred_ &p) const
	{ return assumed_head_pred_<typename Ac_::Inverse_, Pred_>(p); }
	template <class Ac_>
	T_ *remove_assumed_head_()
	{
		T_ *ret = Ac_::get_head_(this);
		if (ret) {
			*Ac_::get_prev_ref_(this, Ac_::get_next_(ret)) = NULL;
			Ac_::set_head_(this, Ac_::get_next_(ret));
			Ac_::clear_node_(ret);
		}
		return ret;
	}
	template <class Ac_>
	T_ *insert_assumed_after_(T_ *base, T_ *n)
	{
		confirm_newnode_("insert_assumed_after_", n);
		T_ **rnext = Ac_::get_next_ref_(this, base);
		*Ac_::get_prev_ref_(this, *rnext) = n;
		Ac_::set_next_(n, *rnext);
		*rnext = n;
		Ac_::set_prev_(n, base);
		return n;
	}
	template <typename Fn_, class Ac_, typename Pred_>
	T_ *each_(const Fn_ &fn, const Pred_ &p) const
	{
		for (T_ *n = this->assumed_head_pred_<Ac_, Pred_>(p), *nxt;
		     n;
		     n=nxt) {
			nxt = this->assumed_next_pred_<Ac_, Pred_>(n, p);
			if (fn(n))
				return n;
		}
		return NULL;
	}
	template <class Ac_>
	void swap_assumed_next_(T_ *base)
	{
		T_ *nxt;
		if ((nxt = Ac_::get_next_(base))) {
			*Ac_::get_next_ref_(this, Ac_::get_prev_(base)) = nxt;
			*Ac_::get_prev_ref_(this, Ac_::get_next_(next)) = base;
			Ac_::set_prev_(next, Ac_::get_prev_(base));
			Ac_::set_next_(base, Ac_::get_next_(nxt));
			Ac_::set_next_(next, base);
			Ac_::set_prev_(base, nxt);
		}
	}
public:
	~Container() { Forward_::dispose(this); }
	Container() : m_head(NULL), m_tail(NULL) { (void)&Forward_::confirm_; }
	bool is_empty() const { return !m_head && !m_tail; }
	T_ *head() const { return m_head; }
	T_ *tail() const { return m_tail; }
	template <typename Pred_>
	T_ *head(const Pred_ &p) const
	{ return this->assumed_head_pred_<Forward_, Pred_>(p); }
	template <typename Pred_>
	T_ *tail(const Pred_ &p) const
	{ return this->assumed_tail_pred_<Forward_, Pred_>(p); }
	T_ *next(T_ *n) const
	{ return n ? Forward_::get_next_(n) : this->head(); }
	T_ *prev(T_ *n) const
	{ return n ? Forward_::get_prev_(n) : this->tail(); }
	template <typename Pred_>
	T_ *next(T_ *n, const Pred_ &p) const
	{
		return n ?
		    this->assumed_next_pred_<Forward_, Pred_>(n, p) :
		    this->head(p);
	}
	template <typename Pred_>
	T_ *prev(T_ *n, const Pred_ &p) const
	{
		return n ?
		    this->assumed_prev_pred_<Forward_, Pred_>(n, p) :
		    this->tail(p);
	}
	void release() { m_head = m_tail = NULL; }
	T_ *remove(T_ *n)
	{
		if (n) {
			AMT_ASSERT_LIBAMT_U_(
				"remove",
				((Forward_::get_next_(n) == NULL) ==
				 (Forward_::get_tail_(this) == n)) &&
				((Forward_::get_prev_(n) == NULL) ==
				 (Forward_::get_head_(this) == n)));
			*Forward_::get_prev_ref_(this, Forward_::get_next_(n))
			    = Forward_::get_prev_(n);
			*Forward_::get_next_ref_(this, Forward_::get_prev_(n))
			    = Forward_::get_next_(n);
			Forward_::clear_node_(n);
		}
		return n;
	}
	T_ *remove_head() { return remove_assumed_head_<Forward_>(); }
	T_ *remove_tail() { return remove_assumed_head_<Reverse_>(); }
	T_ *insert_after(T_ *base, T_ *n)
	{ return insert_assumed_after_<Forward_>(base, n); }
	T_ *insert_before(T_ *base, T_ *n)
	{ return insert_assumed_after_<Reverse_>(base, n); }
	T_ *insert_head(T_ *n) { return insert_after(NULL, n); }
	T_ *insert_tail(T_ *n) { return insert_before(NULL, n); }
	template <typename Fn_>
	T_ *each(const Fn_ &fn) const
	{ return each_<Fn_, Forward_, Predicates::AlwaysTrue>(
		fn, Predicates::AlwaysTrue()); }
	template <typename Fn_>
	T_ *each_reverse(const Fn_ &fn) const
	{ return each_<Fn_, Reverse_, Predicates::AlwaysTrue>(
		fn, Predicates::AlwaysTrue()); }
	template <typename Fn_, typename Pred_>
	T_ *each(const Fn_ &fn, const Pred_ &p) const
	{ return each_<Fn_, Forward_, Pred_>(fn, p); }
	template <typename Fn_, typename Pred_>
	T_ *each_reverse(const Fn_ &fn, const Pred_ &p) const
	{ return each_<Fn_, Backward_, Pred_>(fn, p); }
	template <typename Fn_>
	void remove_all(const Fn_ &fn)
	{
		T_ *t;
		while ((t = this->remove_head()))
			fn(t);
	}
	void swap_next(T_ *base) { swap_assumed_next_<Forward_>(base); }
	void swap_prev(T_ *base) { swap_assumed_next_<Reverse_>(base); }
	void swap(T_ *n1, T_ *n2)
	{
		T_ *n1n = Forward_::get_next_(n1);
		T_ *n1p = Forward_::get_prev_(n1);
		T_ *n2n = Forward_::get_next_(n2);
		T_ *n2p = Forward_::get_prev_(n2);
		Forward_::set_next_(n1, n2n);
		Forward_::set_prev_(n1, n2p);
		*Forward_::get_next_ref_(this, n2p) = n1;
		*Forward_::get_prev_ref_(this, n2n) = n1;
		Forward_::set_next_(n2, n1n);
		Forward_::set_prev_(n2, n1p);
		*Forward_::get_next_ref_(this, n1p) = n2;
		*Forward_::get_prev_ref_(this, n1n) = n2;
	}
};

template <class T_, class A_ = SimpleAccessor<T_> >
class Node
{
	AMT_NOCOPY(Node);
	friend class Container<T_, A_>::Access_;
public:
	typedef A_ Accessor;
	typedef Container<T_, A_> Container;
private:
	T_ *m_next, *m_prev;
	AMT_DEBUG_CODE(static const Integer::UIntPtr deadbeef = 0xDEADBEEF);
	AMT_DEBUG_CODE(bool is_deadbeef() const
	{ return m_next == (T_ *)deadbeef || m_prev == (T_ *)deadbeef; })
public:
	Node() : m_next(NULL), m_prev(NULL) { }
	bool is_sole() const { return !m_next && !m_prev; }
};

}

}

#endif
