#ifndef AMT_WEAK_REF_H_
#define AMT_WEAK_REF_H_

#include <amt/Defines.h>
#include <amt/Types.h>
#include <amt/Assert.h>
#include <amt/ScopedPtr.h>

namespace AMT
{

namespace WeakRefs
{

template <class> class Manager;

template <class, class> class ScopedPtr;

struct Paired
{
	template <class SC_, class Obj_>
	struct ScopedDisposer
	{
		static void dispose(SC_ &sc, Obj_ *obj)
		{ sc.manager_()->weak_unref(obj); }
	};
};

struct Self
{
	template <class SC_, class Obj_>
	struct ScopedDisposer
	{
		static void dispose(SC_ &, Obj_ *obj)
		{ obj->weak_unref(); }
	};
};

struct SimpleHeapDisposer
{
	template <class SC_, class Obj_>
	struct HeapDisposer
	{
		static void dispose(SC_ &, Obj_ *obj)
		{ delete obj; }
	};
};

template <class Traits_>
class ScopedPtr<Traits_, Paired>
	: Ptr::ScopedBase<ScopedPtr<Traits_, Paired>,
			  typename Traits_::Object,
			  Paired::ScopedDisposer<ScopedPtr<Traits_, Paired>,
						 typename Traits_::Object> >
{
	friend class Manager<Traits_>;
	typedef ScopedPtr Self_;
	typedef typename Traits_::RawInheritedManager RIM_;
	typedef typename Traits_::Object Obj_;
	typedef typename Paired::ScopedDisposer<Self_, Obj_> SD_;
	typedef Ptr::ScopedBase<Self_, Obj_, SD_> Super_;
	friend class Ptr::ScopedBase<Self_, Obj_, SD_>;
	friend class Paired::ScopedDisposer<Self_, Obj_>;
	RIM_ *m_manager;
	AMT_PRIV_ACSOR_R(RIM_ *, manager);
public:
	ScopedPtr() : m_manager(NULL) { }
	ScopedPtr(RIM_ &m, Obj_ *obj, bool noincr=false)
		: Super_(obj), m_manager(obj ? &m : NULL)
	{
		if (obj && !noincr) m.weak_ref(obj);
	}
	void reset(RIM_ &m, Obj_ *obj, bool noincr=false)
	{
		this->Super_::reset(obj);
		m_manager = obj ? &m:NULL;
		if (obj && !noincr) m.weak_ref(obj);
	}
	void reset() { this->Super_::reset(); m_manager = NULL; }
	Obj_ *renounce()
	{ m_manager = NULL; return this->Super_::renounce(); }
	using Super_::is_valid;
	using Super_::operator *;
	using Super_::operator ->;
	using Super_::operator [];
	AMT_PTR_TO_REF_ACSOR_CONST(RIM_, manager);
};

template <class Traits_>
class ScopedPtr<Traits_, Self>
	: Ptr::ScopedBase<ScopedPtr<Traits_, Self>,
			  typename Traits_::Object,
			  Self::ScopedDisposer<ScopedPtr<Traits_, Self>,
					       typename Traits_::Object> >
{
	friend class Manager<Traits_>;
	typedef Manager<Traits_> Man_;
	typedef ScopedPtr Self_;
	typedef typename Traits_::Object Obj_;
	typedef Self::ScopedDisposer<Self_, Obj_> SD_;
	typedef Ptr::ScopedBase<Self_, Obj_, SD_> Super_;
	friend class Ptr::ScopedBase<Self_, Obj_, SD_>;
	friend class Self::ScopedDisposer<Self_, Obj_>;
public:
	ScopedPtr(const Man_ &, Obj_ *obj, bool noincr=false)
		: Super_(obj)
	{ if (obj && !noincr) obj->weak_ref(); }
	void reset(const Man_ &, Obj_ *obj, bool noincr=false)
	{ this->Super_::reset(obj); if (obj && !noincr) obj->weak_ref(); }
	void reset() { this->Super_::reset(); }
	using Super_::renounce;
	using Super_::is_valid;
	using Super_::operator *;
	using Super_::operator ->;
	using Super_::operator [];
};

class SimpleCounter
{
	template <class T_> friend class Manager;
	U32 m_weak_ref_count;
	static const U32 COUNT_MASK_ = 0x7FFFFFFF;
	static const U32 ORPHANED_MASK_ = 0x80000000;
	// export for Manager
	void incr_weak_ref_count()
	{ m_weak_ref_count++; }
	void decr_weak_ref_count()
	{ m_weak_ref_count--; }
	void set_orphaned()
	{ m_weak_ref_count |= ORPHANED_MASK_; }
	void clear_orphaned()
	{ m_weak_ref_count &= ~ORPHANED_MASK_; }
public:
	SimpleCounter() : m_weak_ref_count(0) { }
	U32 weak_ref_count() const { return m_weak_ref_count & COUNT_MASK_; }
	bool is_orphaned() const { return m_weak_ref_count & ORPHANED_MASK_; }
	bool is_unused() const { return m_weak_ref_count == ORPHANED_MASK_; }
};

template <class Traits_>
class Manager
{
	typedef typename Traits_::RawInheritedManager RIM_;
	typedef typename Traits_::InheritedManager IM_;
	typedef typename Traits_::Object Object_;
public:
	typedef Object_ ObjectType;
	typedef ScopedPtr<Traits_, typename Traits_::ScopedDisposer> ScopedPtr;
protected:
	~Manager() { }
	Manager() { }
	IM_ &to_im_() { return static_cast<IM_ &>(*this); }
	const IM_ &to_im_() const { return static_cast<const IM_ &>(*this); }
	static void orphaned_(RIM_ &im, Object_ *obj)
	{
		AMT_ASSERT_LIBAMT_U_("orphaned", !obj->is_orphaned());
		Traits_::hook_orphaned(im, obj);
		obj->set_orphaned();
		if (obj->weak_ref_count() == 0)
			Traits_::heap_dispose(im, obj);
	}
	static void reclaim_(RIM_ &im, Object_ *obj)
	{
		AMT_ASSERT_LIBAMT_U_("reclaim", obj->is_orphaned());
		Traits_::hook_reclaim(im, obj);
		obj->clear_orphaned();
	}
	static Object_ *weak_ref_(RIM_ &im, Object_ *obj)
	{
		AMT_ASSERT_LIBAMT_U_("weak_ref", !obj->is_orphaned());
		Traits_::hook_weak_ref(im, obj);
		obj->incr_weak_ref_count();
		return obj;
	}
	static void weak_unref_(RIM_ &im, Object_ *obj)
	{
		AMT_ASSERT_LIBAMT_U_("weak_unref", obj->weak_ref_count()>0);
		Traits_::hook_weak_unref(im, obj);
		obj->decr_weak_ref_count();
		if (obj->is_unused())
			Traits_::heap_dispose(im, obj);
	}
};

template <class IM_, class Obj_, class SD_ = Paired,
	  class HD_ = SimpleHeapDisposer>
struct SimpleTraits
{
	typedef SD_ ScopedDisposer;
	typedef IM_ RawInheritedManager;
	typedef typename Const<IM_>::Deconst InheritedManager;
	typedef Obj_ Object;
	static void hook_orphaned(IM_ &, Obj_ *) { }
	static void hook_weak_ref(IM_ &, Obj_ *) { }
	static void hook_weak_unref(IM_ &, Obj_ *) { }
	static void hook_reclaim(IM_ &, Obj_ *) { }
	static void heap_dispose(IM_ &im, Obj_ *obj)
	{ HD_::template HeapDisposer<IM_, Obj_>::dispose(im, obj); }
};

namespace Bits_
{

template <class, class> struct MTHelper;

template <class Traits_, class IM_>
struct MTHelper<Traits_, const IM_> : public Manager<Traits_>
{
	typedef Manager<Traits_> Super_;
	typedef typename Super_::ObjectType Object_;
public:
	~MTHelper() { }
	MTHelper() { }
	void orphaned(Object_ *obj) const
	{ this->orphaned_(this->to_im_(), obj); }
	Object_ *weak_ref(Object_ *obj) const
	{ return this->weak_ref_(this->to_im_(), obj); }
	void weak_unref(Object_ *obj) const
	{ this->weak_unref_(this->to_im_(), obj); }
	void reclaim(Object_ *obj) const
	{ this->reclaim_(this->to_im_(), obj); }
};

template <class Traits_, class IM_>
struct MTHelper : public MTHelper<Traits_, const IM_>
{
	typedef MTHelper<Traits_, const IM_> Super_;
	typedef typename Super_::ObjectType Object_;
public:
	void orphaned(Object_ *obj)
	{ this->orphaned_(this->to_im_(), obj); }
	Object_ *weak_ref(Object_ *obj)
	{ return this->weak_ref_(this->to_im_(), obj); }
	void weak_unref(Object_ *obj)
	{ this->weak_unref_(this->to_im_(), obj); }
	void reclaim(Object_ *obj)
	{ this->reclaim_(this->to_im_(), obj); }
};

}

template <class Traits_>
class ManagerTmpl : Bits_::MTHelper<Traits_,
				    typename Traits_::RawInheritedManager>
{
	friend class Manager<Traits_>;
	typedef typename Traits_::RawInheritedManager RIM_;
	typedef Bits_::MTHelper<Traits_, RIM_> Super_;
public:
	using Super_::orphaned;
	using Super_::weak_ref;
	using Super_::weak_unref;
	using Super_::reclaim;
	using Super_::ScopedPtr;
};

template <class IM_, class Obj_, class SD_ = Paired,
	  class HD_ = SimpleHeapDisposer>
class SimpleManagerTmpl : ManagerTmpl<SimpleTraits<IM_, Obj_, SD_, HD_> >
{
public:
	typedef SimpleTraits<IM_, Obj_, SD_, HD_> Traits;
private:
	friend class Manager<Traits>;
	typedef ManagerTmpl<Traits> Super_;
public:
	using Super_::orphaned;
	using Super_::weak_ref;
	using Super_::weak_unref;
	using Super_::reclaim;
	using Super_::ScopedPtr;
};

}

}

#endif
